/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2015                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>
#include "SmallObjAllocator.h"

#include <vector>

#include "BlockAllocator.h"
#include "Threading.h"

using namespace std;
using namespace pearl;


//--- Private data structures -----------------------------------------------

struct SmallObjAllocator::tAllocator
{
    /// Container type for storing thread-local allocators
    typedef std::vector<SmallObjAllocator::alloc_container*> threadAllocContainer;


    /// Underlying per-thread allocators
    threadAllocContainer mThread;

    /// Lock to protect allocator container
    Mutex mLock;
};


//---------------------------------------------------------------------------
//
//  class SmallObjAllocator (multi-threaded code)
//
//---------------------------------------------------------------------------

//--- Constructors & destructor ---------------------------------------------

SmallObjAllocator::SmallObjAllocator()
{
  // SmallObjAllocator is a singleton, i.e., the one and only instance
  // can initialize the specific allocator data
  assert(!mAllocator);
  mAllocator = new SmallObjAllocator::tAllocator;

  // Create allocator for master thread
  mAllocator->mThread.push_back(new alloc_container);
}


SmallObjAllocator::~SmallObjAllocator()
{
  // Release allocators
  tAllocator::threadAllocContainer::iterator tit = mAllocator->mThread.begin();
  while (tit != mAllocator->mThread.end())
  {
    alloc_container::iterator it = (*tit)->begin();
    while (it != (*tit)->end()) {
      delete it->second;
      ++it;
    }

    delete *tit;
    ++tit;
  }

  // Release allocator data
  delete mAllocator;
}


//--- Memory allocation routines --------------------------------------------

void* SmallObjAllocator::alloc(std::size_t objectSize) const throw(bad_alloc)
{
  int  threadId = PEARL_GetThreadNumber();
  bool error    = false;

  // Retrieve thread-local allocator
  alloc_container* container;
  mAllocator->mLock.lock();
  // Create missing allocators if necessary
  if (threadId >= mAllocator->mThread.size())
  {
    // We need to catch 'bad_alloc' exceptions that might be thrown by
    // the push_back() and operator new() below and re-throw them outside
    // the critical section.
    try {
      int numThreads = PEARL_GetNumThreads();
      while (mAllocator->mThread.size() < numThreads) {
        mAllocator->mThread.push_back(new alloc_container);
      }
    }
    catch (bad_alloc&) {
      error = true;
    }
  }
  container = mAllocator->mThread[threadId];
  mAllocator->mLock.unlock();

  // Re-throw 'bad_alloc' exception if necessary
  if (error)
    throw bad_alloc();

  // Allocate object
  return SmallObjAllocator::alloc(*container, objectSize);
}


void SmallObjAllocator::dealloc(void*       deadObject,
                                std::size_t objectSize) const
{
  int threadId = PEARL_GetThreadNumber(); 

  // Retrieve thread-local allocator
  mAllocator->mLock.lock();
  assert(threadId < mAllocator->mThread.size());
  alloc_container* container = mAllocator->mThread[threadId];
  mAllocator->mLock.unlock();

  // At this point, we know that the thread-local allocator is in place
  SmallObjAllocator::dealloc(*container, deadObject, objectSize);
}
